# coding: utf-8

"""
    LAMP Platform

    The LAMP Platform API.

    The version of the OpenAPI document: 1.0.0
    Contact: team@digitalpsych.org
"""


from __future__ import absolute_import

import unittest

import LAMP
from LAMP.api.credential_api import CredentialApi
from LAMP.rest import ApiException


class TestCredentialApi(unittest.TestCase):
    """CredentialApi unit test stubs"""

    def setUp(self):
        self.api = LAMP.api.credential_api.CredentialApi()

    def tearDown(self):
        pass

    def test_create(self):
        """Test case for create

        """
        pass

    def test_delete(self):
        """Test case for delete

        """
        pass

    def test_list(self):
        """Test case for list

        """
        pass

    def test_update(self):
        """Test case for update

        """
        pass


if __name__ == '__main__':
    unittest.main()
